/*
 * Perl module for Ruby (Ruby interface for Perl module)
 * 1998 by yoshidam
 */

#include <ruby.h>
#include <st.h>
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#define PERL_POLLUTE 1
#include "embedvar.h"

/* For the earlier versions of Perl */
#ifndef ERRSV
#  define ERRSV GvSV(errgv)
#endif
#ifndef PL_na
#  define PL_na na
#endif
#ifndef PL_sv_undef
#  ifdef sv_undef
#    define PL_sv_undef sv_undef
#  endif
#endif

static VALUE ruby__Sv2Object(SV* sv);
static SV* ruby__Object2Sv(VALUE val);

/* This function originally comes from gc.c of Ruby */
static VALUE
id2ref(char* s_id)
{
  VALUE id = rb_str2inum(s_id, 10);
  unsigned long ptr = NUM2UINT(id);

  if (FIXNUM_P(ptr)) return (VALUE)ptr;
  if (ptr == Qtrue) return Qtrue;
  if (ptr == Qfalse) return Qfalse;
  if (ptr == Qnil) return Qnil;

  ptr = id ^ FIXNUM_FLAG;
#if 0
  if (!looks_pointerp(ptr)) {
    rb_raise(rb_eIndexError, "0x%x is not id value", ptr);
  }
  if (BUILTIN_TYPE(ptr) == 0) {
    rb_raise(rb_eIndexError, "0x%x is recycled object", ptr);
  }
#endif
  return (VALUE)ptr;
}

/* Create Ruby value from Perl value */
static VALUE
ruby__Sv2Object(SV* sv)
{
  int type;
  VALUE obj;
  int len;
  char* ptr;

  if (sv == NULL)
    return Qnil;

  if (SvROK(sv)) {
    SV* av = SvRV(sv);
    if (SvTYPE(av) == SVt_PVAV) {
      VALUE ret = rb_ary_new();
      int len = av_len((AV*)av);
      int i;

      for (i = 0; i <= len; i++) {
	rb_ary_push(ret, ruby__Sv2Object(*av_fetch((AV*)av, i, 0)));
      }
      return ret;
    }
    else if (SvTYPE(av) == SVt_PVHV) {
      VALUE ret = rb_hash_new();
      HE* he;
      
      for (hv_iterinit((HV*)av); he = hv_iternext((HV*)av);) {
	char* key;
	I32 klen;
	SV* val;
	key = hv_iterkey(he, &klen);
	val = hv_iterval((HV*)av, he);
	rb_hash_aset(ret, rb_str_new(key, klen), ruby__Sv2Object(val));
      }
      return ret;
    }
  }


  type = SvTYPE(sv);
  switch (type) {
  case SVt_NULL:
    {
      return Qnil;
    }
  case SVt_IV:
    {
      return INT2NUM(SvIV(sv));
    }
  case SVt_NV:
    {
      return rb_float_new(SvNV(sv));
    }
#if 0
  case SVt_PV:
    {
      ptr = SvPV(sv, len);
      return rb_str_new(ptr, len);
    }
#endif
  }

  /* All other type objects are converted into the string */
  ptr = SvPV(sv, len);
  return rb_str_new(ptr, len);
}

static enum st_retval
__hash_iter(VALUE key, VALUE val, HV* hv)
{
  /* Perl hash key must be a string */
  VALUE vkey = rb_obj_as_string(key);
  hv_store(hv, RSTRING(vkey)->ptr, RSTRING(vkey)->len,
	   ruby__Object2Sv(val), 0);
  return ST_CONTINUE;
}

/* Create Perl value from Ruby value */
static SV*
ruby__Object2Sv(VALUE val)
{
  int type = TYPE(val);

  switch (type) {
  case T_NIL:
    {
      return &PL_sv_undef;
    }
  case T_FIXNUM:
    {
      return newSViv(FIX2INT(val));
    }
  case T_BIGNUM:
    {
      val = rb_obj_as_string(val);
      return newSVpv(RSTRING(val)->ptr, RSTRING(val)->len);
    }
  case T_FLOAT:
    {
      return newSVnv(RFLOAT(val)->value);
    }
  case T_STRING:
    {
      return newSVpv(RSTRING(val)->ptr, RSTRING(val)->len);
    }
  case T_ARRAY:
    {
      AV* av = newAV();
      int len = RARRAY(val)->len;
      int i;
      for (i = 0; i < len; i++) {
	av_push(av, ruby__Object2Sv(RARRAY(val)->ptr[i]));
      }
      return newRV_inc((SV*)av);
    }
  case T_HASH:
    {
      HV* hv = newHV();
      st_table* table = RHASH(val)->tbl;
      st_foreach(table, __hash_iter, hv);
      return newRV_inc((SV*)hv);
    }
  }

#if 0
  /* All other type objects are converted into the string */
  val = rb_obj_as_string(val);
  return newSVpv(RSTRING(val)->ptr, RSTRING(val)->len);
#else
  /* All other type objects are converted into id value */
  val =  rb_obj_as_string(INT2NUM((long)val));
  return newSVpv(RSTRING(val)->ptr, RSTRING(val)->len);
#endif
}

MODULE = Ruby	PACKAGE = Ruby
PROTOTYPES: ENABLE

SV*
eval(str)
    char* str
CODE:
{
  VALUE ret;
  ret = rb_eval_string(str);
  RETVAL = ruby__Object2Sv(ret);
}
OUTPUT:
  RETVAL

SV*
call(func, ...)
    char* func
CODE:
{
  VALUE ret;
  int i;
  VALUE args[16];

  if (items > 16)
    croak("Too many Ruby args");

  for (i = 1; i < items; i++) {
    args[i-1] = ruby__Sv2Object(ST(i));
  }
  ret = rb_funcall2(rb_cObject, rb_intern(func), items-1, args);
  RETVAL = ruby__Object2Sv(ret);
}
OUTPUT:
  RETVAL

SV*
send(oid, method, ...)
    char* oid
    char* method
CODE:
{
  VALUE ret;
  int i;
  VALUE args[16];
  VALUE obj;

  if (items > 16)
    croak("Too many Ruby args");

  for (i = 2; i < items; i++) {
    args[i-2] = ruby__Sv2Object(ST(i));
  }
  ret = rb_funcall2(id2ref(oid), rb_intern(method), items-2, args);
  RETVAL = ruby__Object2Sv(ret);
}
OUTPUT:
  RETVAL
